import typing as t

from pyec.maths import miller_rabin, modular_inverse


class AffinePoint:
    """
    Represents an elliptic curve point in Affine coordinates: (x (mod m), y (mod m))
    """

    def __init__(self, x: int, y: int, m: int) -> None:
        """
        Initializes a point in Affine coordinates from two integers

        Parameters:
            x (int): The first coordinate of the point
            y (int): The second coordinate of the point
            m (int): The modulus of the coordinates

        Returns:
            None
        """
        self.x = x % m
        self.y = y % m
        self.m = m

    def __repr__(self) -> str:
        return f"({self.x}, {self.y})"

    def __getitem__(self, index: int) -> int:
        """
        Enables 0-indexing of an Affine point

        Parameters:
            index (int): Specifies which coordinate of the point to return

        Returns:
            int: The coordinate corresponding to the given index

        Raises:
            IndexError: If the given index is neither 0 or 1
        """
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        else:
            raise IndexError("Index out of range. Affine points have two coordinates.")

    def __eq__(self, other: t.Any) -> bool:
        """
        Enables comparisons between instances of an Affine point and other curve point types

        Parameters:
            other (Any): The object to compare to this point

        Returns:
            bool: True if this and the compared point represent the same point, False if not

        Raises:
            TypeError: If the compared object does not represent a point (in Affine or Jacobian coordinates or infinity point)
        """
        if isinstance(other, AffinePoint):
            return (self[0] == other[0]) and (self[1] == other[1])
        elif isinstance(other, JacobianPoint):
            return self == other.to_affine()
        elif isinstance(other, Infinity):
            return False
        else:
            raise TypeError(f"{other} is not a point.")

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def negate(self) -> "AffinePoint":
        return self.__class__(self.x, -self.y % self.m, self.m)

    def to_affine(self) -> "AffinePoint":
        return self

    def to_jacobian(self) -> "JacobianPoint":
        """
        Converts this point to an equivalent point represented in Jacobian coordinates.
        An Affine point (x, y) is equivalent to the Jacobian point (x, y, 1).

        Parameters:
            None

        Returns:
            JacobianPoint: The point represented in Jacobian coordinates.
        """
        return JacobianPoint(self.x, self.y, 1, self.m)


class JacobianPoint:
    """
    Represents an elliptic curve point in Jacobian coordinates: (x (mod m), y (mod m), z (mod m))
    """

    def __init__(self, x: int, y: int, z: int, m: int) -> None:
        """
        Initializes a point in Jacobian coordinates from three integers

        Parameters:
            x (int): The first coordinate of the point
            y (int): The second coordinate of the point
            z (int): The third coordinate of the point
            m (int): The modulus of the coordinates

        Returns:
            None
        """
        self.x = x % m
        self.y = y % m
        self.z = z % m
        self.m = m

    def __repr__(self) -> str:
        return f"({self.x}, {self.y}, {self.z})"

    def __getitem__(self, index: int) -> int:
        """
        Enables 0-indexing of a Jacobian point

        Parameters:
            index (int): Specifies which coordinate of the point to return, where
            a value of 0 corresponds to the x coordinate, a value of 1 corresponds
            to the y coordinate, and a value of 2 corresponds to the z coordinate

        Returns:
            int: The coordinate corresponding to the given index

        Raises:
            IndexError: If the given index is neither 0, 1, or 2
        """
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        elif index == 2:
            return self.z
        else:
            raise IndexError(
                "Index out of range. Jacobian points have three coordinates."
            )

    def __eq__(self, other: t.Any) -> bool:
        """
        Enables comparisons between instances of a Jacobian point and other curve point types

        Parameters:
            other (Any): The object to compare to this point

        Returns:
            bool: True if this and the compared point represent the same point, False if not

        Raises:
            TypeError: If the compared object does not represent a point (in Affine or Jacobian coordinates or infinity point)
        """
        if isinstance(other, JacobianPoint):
            return (
                (self[0] == other[0])
                and (self[1] == other[1])
                and (self[2] == other[2])
            )
        elif isinstance(other, AffinePoint):
            return self == other.to_jacobian()
        elif isinstance(other, Infinity):
            return False
        else:
            raise TypeError(f"{other} is not a point.")

    def __hash__(self) -> int:
        return hash((self.x, self.y, self.z))

    def negate(self) -> "JacobianPoint":
        return self.__class__(self.x, -self.y % self.m, self.z, self.m)

    def to_affine(self) -> "AffinePoint":
        """
        Converts this point to an equivalent point represented in Affine coordinates.
        A Jacobian point (x, y, z) is equivalent to the Jacobian point (x/z^2, y/z^3).

        Parameters:
            None

        Returns:
            AffinePoint: The point represented in Affine coordinates
        """
        Zinv = modular_inverse(self.z, self.m)
        Zinv2 = Zinv**2
        return AffinePoint(self.x * Zinv2, self.y * (Zinv2 * Zinv), self.m)

    def to_jacobian(self) -> "JacobianPoint":
        return self


class Infinity:
    """
    Represents the infinity point of an elliptic curve
    """

    def __repr__(self) -> str:
        return "Infinity"

    def __eq__(self, other: t.Any) -> bool:
        return isinstance(other, Infinity)

    def negate(self) -> "Infinity":
        return self

    def __hash__(self) -> int:
        return hash(str(self))

    def to_affine(self) -> "Infinity":
        return self

    def to_jacobian(self) -> "Infinity":
        return self



import random
import typing as t


class FiniteField:
    """
    Represents a finite field of odd, prime order
    """

    def __init__(self, p: int) -> None:
        """
        Initializes a finite field instance of the given order

        Parameters:
            p (int): The order of finite field

        Returns:
            None

        Raises:
            ValueError: If the given order is not an odd prime
        """
        if not (miller_rabin(p) and p != 2):
            raise ValueError("Invalid field parameter.")
        self.p = p

    def __repr__(self) -> str:
        return f"GF({self.p})"

    def __contains__(self, res: int) -> bool:
        """
        Checks whether a given integer is an element of the finite field.

        Parameters:
            res (int): The integer to check membership for

        Returns:
            bool: True if the integer is in the field, False if not.
        """
        return res >= 0 and res < self.p

    def __len__(self) -> int:
        return self.p

    class FiniteFieldIterator(t.Iterator[int]):
        """
        Implements the __iter__ and __next__ methods to enable
        in-order iteration through instances of FiniteField
        """

        def __init__(self, p: int) -> None:
            self.p = p
            self.current_value = 0

        def __iter__(self) -> t.Iterator[int]:
            return self

        def __next__(self) -> int:
            if self.current_value < self.p:
                val = self.current_value
                self.current_value += 1
                return val
            else:
                raise StopIteration

    def __iter__(self) -> "FiniteFieldIterator":
        return self.FiniteFieldIterator(self.p)


def miller_rabin(n: int, k: int = 5) -> bool:
    """
    Performs the Miller-Rabin primality test to determine (with high probability)
    if the given integer is prime.

    Parameters:
        n (int): The integer to test for primality
        k (int): The number of trials of the test, note that more trials guarantees
        a lower probability of false positives

    Returns:
        bool: True if the integer is prime (or composite with probability 1/4^k),
        False otherwise.
    """
    if n in (2, 3):
        return True
    if n <= 1 or n % 2 == 0:
        return False

    s, d = 0, n - 1
    while d % 2 == 0:
        s += 1
        d //= 2

    for _ in range(k):
        a = random.randint(2, n - 2)
        if pow(a, d, n) != 1:
            for i in range(s):
                if pow(a, (2**i) * d, n) == n - 1:
                    break
            else:
                return False

    return True


def extended_euclidean(a: int, b: int) -> t.Tuple[int, int, int]:
    """
    Computes the greatest common divisor (gcd) and Bézout coefficients for the
    two given integers using the extended Euclidean algorithm

    Parameters:
        a (int): The first input to the extended Euclidean algorithm
        b (int): The second input to the extended Euclidean algorithm

    Returns:
        Tuple[int, int, int]: A triplet (x, y, z) such that x = ay + bz

    Raises:
        ValueError: if a = b = 0, as the greatest common divisor of 0 and 0
        is undefined
    """
    if not (a and b):
        if not (a or b):
            raise ValueError("gcd(0, 0) is not defined")
        return (a or b, 0 if b else 1, 0 if a else 1)

    r_old, r = a, b
    s_old, s = 1, 0
    t_old, t = 0, 1

    while r:
        q = r_old // r
        r_old, r = r, r_old - (q * r)
        s_old, s = s, s_old - (q * s)
        t_old, t = t, t_old - (q * t)

    return (r_old, s_old, t_old)


def modular_inverse(a: int, m: int) -> int:
    """
    Computes the multiplicative inverse of some value given a modulus
    using the extended Euclidean algorithm

    Parameters:
        a (int): The value to find the modular multiplicative inverse of
        m (int): The modulus

    Returns:
        int: An integer b such that ab = 1 (mod m)

    Raises:
        ValueError: If the given value has no multiplicative inverse in the given modulus
    """
    bezout_coefs = extended_euclidean(a, m)
    if not bezout_coefs[0] == 1:
        raise ValueError(f"{a} has no multiplicative inverse modulo {m}.")
    else:
        return bezout_coefs[1] % m


def to_binary(n: int) -> t.List[int]:
    """
    Returns the bit representation of a given integer

    Parameters:
        n (int): The integer to convert to binary

    Returns:
        List: A list containing only 0 or 1 representing the
        binary form of the given integer with the most significant
        bit on the left
    """
    if not n:
        return [0]
    bits: t.List[int] = []
    n = abs(n)
    while n:
        bits.insert(0, n % 2)
        n = n // 2
    return bits


def to_naf(n: int, w: int = 2) -> t.List[int]:
    """
    Returns the w-ary non-adjacent representation of a given integer

    Parameters:
        n (int): The integer to convert
        w (int): Determines the range of allowable non-zero values in the non-adjacent form

    Returns:
        List: A list of integers in the w-ary non-adjacent form with the least significant bit on the left
    """
    if not n:
        return [0]

    def mods(a: int, b: int) -> int:
        if (a % b) >= b // 2:
            return (a % b) - b
        else:
            return a % b

    naf: t.List[int] = []
    while n:
        if n % 2:
            z = mods(n, 2**w)
            naf.append(z)
            n -= z
        else:
            naf.append(0)
        n = n // 2

    return naf



"""benchmark.py

The purpose of this module is to benchmark the speed of this implementation of
elliptic curve arithmetic by returning the average time it takes to generate a
key pair, signature, and verification for each of the curves defined in curve_params.py.

Run the module using `python -m pyec.benchmark` and the average time (in milliseconds) for
each major operation in ECDSA for each curve will be printed to the console using the
`tabulate` package.

Global variables PRECISION and TRIALS can be changed to modify experimental parameters.
"""

import statistics
import time
import typing as t

from tabulate import tabulate

from pyec.curve_params import REGISTRY
from pyec.point import AffinePoint
from pyec.sign import CurveSign, Signature

PRECISION: int = 4
TRIALS: int = 100


def benchmark_key_generation(signer: CurveSign, trials: int = TRIALS) -> float:
    times = []
    for _ in range(trials):
        start = time.time()
        signer.generate_key_pair()
        end = time.time()
        times.append(end - start)
    return round(statistics.fmean(times), PRECISION) * 10**3


def benchmark_sign(
    message: str, priv_key: int, signer: CurveSign, trials: int = TRIALS
) -> float:
    times = []
    for _ in range(trials):
        start = time.time()
        signature = signer.sign(message, priv_key)
        end = time.time()
        times.append(end - start)
    return round(statistics.fmean(times), PRECISION) * 10**3


def benchmark_verify(
    message: str,
    signature: Signature,
    pub_key: AffinePoint,
    signer: CurveSign,
    trials: int = TRIALS,
) -> float:
    times = []
    for _ in range(trials):
        start = time.time()
        assert signer.verify(message, signature, pub_key)
        end = time.time()
        times.append(end - start)
    return round(statistics.fmean(times), PRECISION) * 10**3


if __name__ == "__main__":
    results = []
    message = "Lorem ipsum dolor sit amet, consectetur adipiscing elit"
    for c in REGISTRY:
        signer = CurveSign(c)
        key_gen_time = benchmark_key_generation(signer)

        key_pair = signer.generate_key_pair()
        priv_key, pub_key = key_pair.priv_key, key_pair.pub_key
        signature = signer.sign(message, priv_key)

        sign_time = benchmark_sign(message, priv_key, signer)
        verify_time = benchmark_verify(message, signature, pub_key, signer)
        results.append([c, key_gen_time, sign_time, verify_time])

    headers = ["Curve", "Key gen (ms)", "Sign (ms)", "Verify (ms)"]
    print(tabulate(results, headers=headers))



import typing as t
from abc import ABC, abstractmethod

from pyec.maths import FiniteField, to_naf
from pyec.point import AffinePoint, Infinity, JacobianPoint

CurveElement = t.Union[AffinePoint, JacobianPoint, Infinity]


class Curve(ABC):
    """
    An interface for the elliptic curve type. This abstract base class is intended
    to subclassed by other classes in this module to implement an elliptic curve
    defined according to a given form. Currently only (short) Weierstrass curves
    are supported.
    """

    def __init__(self, a: int, b: int, p: int) -> None:
        """
        Initializes an elliptic curve with the given parameters with coordinates in the field GF(p).

        Parameters:
            a (int): The first coefficient of the curve equation
            b (int): The second coefficient of the curve equation
            p (int): The prime defining the curve's base field

        Returns:
            None

        Raises:
            ValueError: If the given parameters do not form a non-singular
            curve or a finite field
        """
        self.a = a % p
        self.b = b % p
        self.p = p
        self.base_field = FiniteField(p)
        self._validate_params()

    @abstractmethod
    def _validate_params(self) -> None:
        """
        Checks that the parameters of the curve define a legal, non-singular elliptic curve.

        Parameters:
            None

        Returns:
            None

        Raises:
            ValueError: If for the given type of curve, the following conditions are met:

                Weierstrass curve: 4a^3 + 27b^2 = 0
        """
        ...

    @property
    def infinity(self) -> Infinity:
        """
        Returns the curve's infinity point, which acts as the identity element
        of the group.

        Parameters:
            None

        Returns:
            Infinity: An instance of the Infinity type representing the curve's
            infinity point.
        """
        return Infinity()

    @property
    @abstractmethod
    def points(self) -> t.List[CurveElement]:
        """
        Computes the list of all points on the curve.

        **Disclaimer: This method is very inefficient and should only be called
        for small curves. In particular, do not invoke this method on any of the
        NIST/Brainpool curves (i.e. the ones defined in curve_params.py)

        Parameters:
            None

        Returns:
            List: A list of all points, including infinity, on the curve.
        """
        ...

    @abstractmethod
    def __repr__(self) -> str:
        ...

    @abstractmethod
    def __contains__(self, P: CurveElement) -> bool:
        """
        Determines if the provided point exists on the curve by checking if the
        point (in Affine form) satisfies the curve equation

        Parameters:
            P (CurveElement): The point to check membership for

        Returns:
            bool: True if the given point is on the curve or if the point is
            the infinity point. False otherwise.
        """
        ...

    @abstractmethod
    def add(
        self, P: CurveElement, Q: CurveElement, to_affine: bool = False
    ) -> CurveElement:
        """
        Computes the sum of two elliptic curve points P and Q.

        Parameters:
            P (CurveElement): The first summand
            Q (CurveElement): The second summand
            to_affine (int): Determine the coordinate system in which to return
            the resulting point, defaults to Jacobian

        Returns:
            CurveElement: The point on the curve R such that R = P + Q
        """
        ...

    @t.overload
    def scalar_mult(
        self,
        P: CurveElement,
        n: int,
        w: int,
        to_affine: t.Literal[True],
    ) -> AffinePoint:
        ...

    @t.overload
    def scalar_mult(
        self,
        P: CurveElement,
        n: int,
        w: int,
        to_affine: t.Literal[False],
    ) -> JacobianPoint:
        ...

    @t.overload
    def scalar_mult(
        self,
        P: CurveElement,
        n: int,
    ) -> AffinePoint:
        ...

    @t.overload
    def scalar_mult(
        self,
        P: CurveElement,
        n: int,
        w: int,
    ) -> AffinePoint:
        ...

    def scalar_mult(
        self, P: CurveElement, n: int, w: int = 2, to_affine: bool = False
    ) -> CurveElement:
        """
        Computes a scalar multiple of the elliptic curve point P using the double-and-add
        algorithm on the w-ary non-adjacent form (wNAF) of the scalar.

        Parameters:
            P (CurveElement): The curve point to scale
            n (int): The integer to scale the curve point by
            w (int): Determines the range of allowable non-zero values in the non-adjacent form
            to_affine (int): Determine the coordinate system in which to return
            the resulting point, defaults to Jacobian

        Returns:
            CurveElement: The point on the curve Q such that Q = nP

        Raises:
            ValueError: If the scalar n is negative (n < 0)
        """
        if n < 0:
            raise ValueError("Power must be non-negative.")
        if isinstance(P, Infinity):
            return P
        bits = to_naf(n, w)
        R: CurveElement = self.infinity

        mults = {i: self._scalar_mult(P, i) for i in range(1, 2 ** (w - 1), 2)}

        for bit in reversed(bits):
            R = self._double(R)
            if bit > 0:
                R = self.add(R, mults[bit])
            elif bit < 0:
                R = self.add(R, mults[bit * -1].negate())

        return R if not to_affine else R.to_affine()

    def _scalar_mult(self, P: CurveElement, n: int) -> CurveElement:
        if n < 0:
            raise ValueError("Power must be non-negative.")
        if isinstance(P, Infinity):
            return P
        bits = to_naf(n)
        Q: CurveElement = P.to_jacobian()
        R: CurveElement = self.infinity
        for bit in bits:
            if bit == 1:
                R = self.add(R, Q)
            elif bit == -1:
                R = self.add(R, Q.negate())
            Q = self._double(Q)

        return R

    @abstractmethod
    def _double(self, P: CurveElement) -> CurveElement:
        """
        Doubles the elliptic curve point P.

        Parameters:
            P (Point): The point to double

        Returns:
            CurveElement: The point on the curve Q such that Q = 2P
        """
        ...

    def create_point(self, x: int, y: int) -> JacobianPoint:
        """
        Creates an elliptic curve point (represented in Jacobian coordinates) with
        the given coordinates.

        Parameters:
            x (int): The x coordinate of the curve to generate
            y (int): The y coordinate of the curve to generate

        Returns:
            JacobianPoint: A point on the curve in Jacobian form (x, y, 1)

        Raises:
            ValueError: If the point (x, y) does not exist on the curve
        """
        point = AffinePoint(x, y, self.p)
        if point not in self:
            raise ValueError("The given point is not on the curve.")
        return point.to_jacobian()


class ShortWCurve(Curve):
    """
    Represents an elliptic curve in (short) Weierstrass form
    given by the equation Y^2 = X^3 + aX + b.
    """

    def _validate_params(self) -> None:
        if ((4 * int(self.a) ** 3) + (27 * int(self.b) ** 2)) % self.p == 0:
            raise ValueError("Discriminant must be non-zero.")

    def __repr__(self) -> str:
        return f"ShortWCurve(a={self.a}, b={self.b}) over {self.base_field}"

    def __contains__(self, P: CurveElement) -> bool:
        if isinstance(P, Infinity):
            return True
        P = P.to_affine()
        return (P[1] ** 2) % self.p == (P[0] ** 3 + self.a * P[0] + self.b) % self.p

    @property
    def points(self) -> t.List[CurveElement]:
        points: t.Set[CurveElement] = set()
        values: t.Dict[str, t.List[int]] = {
            "input": [],
            "lhs": [],
            "rhs": [],
        }
        for res in self.base_field:
            values["input"].append(res)
            values["lhs"].append(res**2)
            values["rhs"].append(res**3 + self.a * res + self.b)
        for i, val in enumerate(values["rhs"]):
            if val in values["lhs"]:
                indices = [j for j, x in enumerate(values["lhs"]) if x == val]
                for j in indices:
                    points.add(
                        AffinePoint(values["input"][i], values["input"][j], self.p)
                    )

        points.add(self.infinity)

        return list(points)

    def add(
        self, P: CurveElement, Q: CurveElement, to_affine: bool = False
    ) -> CurveElement:
        if isinstance(P, Infinity):
            return Q if not to_affine else Q.to_affine()
        elif isinstance(Q, Infinity):
            return P if not to_affine else P.to_affine()
        elif P.negate() == Q:
            return self.infinity
        else:
            if P == Q:
                return self._double(P)
            else:
                X_1, Y_1, Z_1 = P[0], P[1], P[2]
                X_2, Y_2, Z_2 = Q[0], Q[1], Q[2]
                U, V = (X_1 * Z_2**2) % self.p, (X_2 * Z_1**2) % self.p
                S, T = (Y_1 * Z_2**3) % self.p, (Y_2 * Z_1**3) % self.p
                if U == V:
                    if S == T:
                        return self._double(P)
                    return self.infinity
                H = (V - U) % self.p
                R = (T - S) % self.p
                H2 = H**2 % self.p
                H3 = (H2 * H) % self.p
                UH2 = (U * H2) % self.p

                x = (R**2 - H3 - (2 * UH2)) % self.p
                y = (R * (UH2 - x) - (S * H3)) % self.p
                z = (H * Z_1 * Z_2) % self.p
                res = JacobianPoint(x, y, z, self.p)

                return res if not to_affine else res.to_affine()

    def _double(self, P: CurveElement) -> CurveElement:
        if isinstance(P, Infinity):
            return P
        elif isinstance(P, AffinePoint):
            P = P.to_jacobian()
        elif P[1] == 0:
            return self.infinity
        X, Y, Z = P[0], P[1], P[2]
        Y2, Z2 = Y**2, Z**2
        S = (4 * X * Y2) % self.p
        if self.a == self.p - 3:
            M = (3 * (X + Z2) * (X - Z2)) % self.p
        else:
            M = (3 * X**2 + int(self.a) * Z2**2) % self.p
        T = (M**2 - 2 * S) % self.p
        x = T
        y = (M * (S - T) - (8 * Y2**2)) % self.p
        z = (2 * Y * Z) % self.p

        return JacobianPoint(x, y, z, self.p)



from collections import namedtuple

from pyec.curve import Curve, ShortWCurve

CurveParams = namedtuple("CurveParams", ["p", "h", "n", "type", "a", "b", "g_x", "g_y"])

REGISTRY = {
    "P-224": CurveParams(
        p=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000001,
        h=0x1,
        n=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFF16A2E0B8F03E13DD29455C5C2A3D,
        type="shortw",
        a=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFE,
        b=0xB4050A850C04B3ABF54132565044B0B7D7BFD8BA270B39432355FFB4,
        g_x=0xB70E0CBD6BB4BF7F321390B94A03C1D356C21122343280D6115C1D21,
        g_y=0xBD376388B5F723FB4C22DFE6CD4375A05A07476444D5819985007E34,
    ),
    "P-256": CurveParams(
        p=0xFFFFFFFF00000001000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF,
        h=0x1,
        n=0xFFFFFFFF00000000FFFFFFFFFFFFFFFFBCE6FAADA7179E84F3B9CAC2FC632551,
        type="shortw",
        a=0xFFFFFFFF00000001000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFC,
        b=0x5AC635D8AA3A93E7B3EBBD55769886BC651D06B0CC53B0F63BCE3C3E27D2604B,
        g_x=0x6B17D1F2E12C4247F8BCE6E563A440F277037D812DEB33A0F4A13945D898C296,
        g_y=0x4FE342E2FE1A7F9B8EE7EB4A7C0F9E162BCE33576B315ECECBB6406837BF51F5,
    ),
    "P-384": CurveParams(
        p=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFFFF0000000000000000FFFFFFFF,
        h=0x1,
        n=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFC7634D81F4372DDF581A0DB248B0A77AECEC196ACCC52973,
        type="shortw",
        a=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFFFF0000000000000000FFFFFFFC,
        b=0xB3312FA7E23EE7E4988E056BE3F82D19181D9C6EFE8141120314088F5013875AC656398D8A2ED19D2A85C8EDD3EC2AEF,
        g_x=0xAA87CA22BE8B05378EB1C71EF320AD746E1D3B628BA79B9859F741E082542A385502F25DBF55296C3A545E3872760AB7,
        g_y=0x3617DE4A96262C6F5D9E98BF9292DC29F8F41DBD289A147CE9DA3113B5F0B8C00A60B1CE1D7E819D7A431D7C90EA0E5F,
    ),
    "P-521": CurveParams(
        p=0x1FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,
        h=0x1,
        n=0x1FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFA51868783BF2F966B7FCC0148F709A5D03BB5C9B8899C47AEBB6FB71E91386409,
        type="shortw",
        a=0x1FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFC,
        b=0x051953EB9618E1C9A1F929A21A0B68540EEA2DA725B99B315F3B8B489918EF109E156193951EC7E937B1652C0BD3BB1BF073573DF883D2C34F1EF451FD46B503F00,
        g_x=0xC6858E06B70404E9CD9E3ECB662395B4429C648139053FB521F828AF606B4D3DBAA14B5E77EFE75928FE1DC127A2FFA8DE3348B3C1856A429BF97E7E31C2E5BD66,
        g_y=0x11839296A789A3BC0045C8A5FB42C7D1BD998F54449579B446817AFBD17273E662C97EE72995EF42640C550B9013FAD0761353C7086A272C24088BE94769FD16650,
    ),
    "brainpoolP224r1": CurveParams(
        p=0xD7C134AA264366862A18302575D1D787B09F075797DA89F57EC8C0FF,
        h=0x1,
        n=0xD7C134AA264366862A18302575D0FB98D116BC4B6DDEBCA3A5A7939F,
        type="shortw",
        a=0x68A5E62CA9CE6C1C299803A6C1530B514E182AD8B0042A59CAD29F43,
        b=0x2580F63CCFE44138870713B1A92369E33E2135D266DBB372386C400B,
        g_x=0x0D9029AD2C7E5CF4340823B2A87DC68C9E4CE3174C1E6EFDEE12C07D,
        g_y=0x58AA56F772C0726F24C6B89E4ECDAC24354B9E99CAA3F6D3761402CD,
    ),
    "brainpoolP256r1": CurveParams(
        p=0xA9FB57DBA1EEA9BC3E660A909D838D726E3BF623D52620282013481D1F6E5377,
        h=0x1,
        n=0xA9FB57DBA1EEA9BC3E660A909D838D718C397AA3B561A6F7901E0E82974856A7,
        type="shortw",
        a=0x7D5A0975FC2C3057EEF67530417AFFE7FB8055C126DC5C6CE94A4B44F330B5D9,
        b=0x26DC5C6CE94A4B44F330B5D9BBD77CBF958416295CF7E1CE6BCCDC18FF8C07B6,
        g_x=0x8BD2AEB9CB7E57CB2C4B482FFC81B7AFB9DE27E1E3BD23C23A4453BD9ACE3262,
        g_y=0x547EF835C3DAC4FD97F8461A14611DC9C27745132DED8E545C1D54C72F046997,
    ),
    "brainpoolP384r1": CurveParams(
        p=0x8CB91E82A3386D280F5D6F7E50E641DF152F7109ED5456B412B1DA197FB71123ACD3A729901D1A71874700133107EC53,
        h=0x1,
        n=0x8CB91E82A3386D280F5D6F7E50E641DF152F7109ED5456B31F166E6CAC0425A7CF3AB6AF6B7FC3103B883202E9046565,
        type="shortw",
        a=0x7BC382C63D8C150C3C72080ACE05AFA0C2BEA28E4FB22787139165EFBA91F90F8AA5814A503AD4EB04A8C7DD22CE2826,
        b=0x04A8C7DD22CE28268B39B55416F0447C2FB77DE107DCD2A62E880EA53EEB62D57CB4390295DBC9943AB78696FA504C11,
        g_x=0x1D1C64F068CF45FFA2A63A81B7C13F6B8847A3E77EF14FE3DB7FCAFE0CBD10E8E826E03436D646AAEF87B2E247D4AF1E,
        g_y=0x8ABE1D7520F9C2A45CB1EB8E95CFD55262B70B29FEEC5864E19C054FF99129280E4646217791811142820341263C5315,
    ),
    "brainpoolP512r1": CurveParams(
        p=0xAADD9DB8DBE9C48B3FD4E6AE33C9FC07CB308DB3B3C9D20ED6639CCA703308717D4D9B009BC66842AECDA12AE6A380E62881FF2F2D82C68528AA6056583A48F3,
        h=0x1,
        n=0xAADD9DB8DBE9C48B3FD4E6AE33C9FC07CB308DB3B3C9D20ED6639CCA70330870553E5C414CA92619418661197FAC10471DB1D381085DDADDB58796829CA90069,
        type="shortw",
        a=0x7830A3318B603B89E2327145AC234CC594CBDD8D3DF91610A83441CAEA9863BC2DED5D5AA8253AA10A2EF1C98B9AC8B57F1117A72BF2C7B9E7C1AC4D77FC94CA,
        b=0x3DF91610A83441CAEA9863BC2DED5D5AA8253AA10A2EF1C98B9AC8B57F1117A72BF2C7B9E7C1AC4D77FC94CADC083E67984050B75EBAE5DD2809BD638016F723,
        g_x=0x81AEE4BDD82ED9645A21322E9C4C6A9385ED9F70B5D916C1B43B62EEF4D0098EFF3B1F78E2D0D48D50D1687B93B97D5F7C6D5047406A5E688B352209BCB9F822,
        g_y=0x7DDE385D566332ECC0EABFA9CF7822FDF209F70024A57B1AA000C55B881F8111B2DCDE494A5F485E5BCA4BD88A2763AED1CA2B2FA8F0540678CD1E0F3AD80892,
    ),
    "secp256k1": CurveParams(
        p=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F,
        h=0x1,
        n=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141,
        type="shortw",
        a=0x0,
        b=0x7,
        g_x=0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798,
        g_y=0x483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8,
    ),
}


def get_curve_params(name: str) -> CurveParams:
    """
    Retrieve the parameters of one of several pre-defined curves suitable
    for use in ECDSA.

    Parameters:
        name (str): Name of curve to get parameters for

    Returns:
        CurveParams: A named tuple containing the following parameters for a given curve

            p: The prime defining the curve's base field
            h: The cofactor of the base point (h = |E| / n)
            n: The order of the base point
            type: The form of the curve
            a: The first coefficient of the curve equation
            b: The second coefficient of the curve equation
            g_x: The first coordinate of the base point
            g_y: The second coordinate of the base point

    Raises:
        ValueError: If the parameters of the requested curve
        are not defined in this module
    """
    return REGISTRY[name]



import hashlib
from random import SystemRandom

from pyec.curve import Curve, ShortWCurve
from pyec.curve_params import CurveParams, get_curve_params
from pyec.maths import modular_inverse
from pyec.point import AffinePoint, Infinity, JacobianPoint

W = 4

class KeyPair:
    pub_key: AffinePoint
    priv_key: int

    def __init__(self, pub_key: AffinePoint, priv_key: int) -> None:
        self.pub_key = pub_key
        self.priv_key = priv_key


class Signature:
    r: int
    s: int

    def __init__(self, r: int, s: int) -> None:
        self.r = r
        self.s = s

    def __str__(self) -> str:
        return f"Signature(r={hex(self.r)}, s={hex(self.s)})"


class CurveSign:
    """
    Extends the base elliptic curve type to support the elliptic curve
    digital signature algorithm (ECDSA)
    """

    params: CurveParams
    curve: Curve
    base_point: JacobianPoint

    def __init__(self, curve_name: str) -> None:
        self.params = get_curve_params(curve_name)
        if self.params.type == "shortw":
            self.curve = ShortWCurve(self.params.a, self.params.b, self.params.p)
        else:
            raise ValueError("Unknown curve type.")
        self.base_point = self.curve.create_point(self.params.g_x, self.params.g_y)

    def _hash(self, m: str) -> int:
        message_len = self.params.n.bit_length()
        if message_len <= 256:
            return int(hashlib.sha256(bytes(m, "utf-8")).hexdigest(), 16)
        elif message_len <= 384:
            return int(hashlib.sha384(bytes(m, "utf-8")).hexdigest(), 16)
        else:
            return int(hashlib.sha512(bytes(m, "utf-8")).hexdigest(), 16)

    def generate_key_pair(self) -> KeyPair:
        d = SystemRandom().randrange(1, self.params.n)
        Q = self.curve.scalar_mult(self.base_point, d, W, to_affine=True)
        return KeyPair(Q.to_affine(), d)

    def sign(self, m: str, priv_key: int) -> Signature:
        h = self._hash(m)
        r, s = 0, 0
        while int(r) == 0:
            k = SystemRandom().randrange(1, self.params.n)
            P = self.curve.scalar_mult(self.base_point, k, W, to_affine=True)
            r = P[0]
        while int(s) == 0:
            s = ((h + priv_key * r) * modular_inverse(k, self.params.n)) % self.params.n
        return Signature(r, s)

    def verify(self, m: str, signature: Signature, pub_key: AffinePoint) -> bool:
        h = self._hash(m)
        r, s = signature.r, signature.s
        if r <= 1 or r > self.params.n or s <= 1 or s > self.params.n:
            return False
        c = modular_inverse(signature.s, self.params.n)
        u, v = (h * c) % self.params.n, (signature.r * c) % self.params.n
        P = self.curve.add(
            self.curve.scalar_mult(self.base_point, u, W),
            self.curve.scalar_mult(pub_key, v, W),
        )
        if isinstance(P, Infinity):
            return False
        return P.to_affine()[0] % self.params.n == r
